# Microsoft C makefile for mawk,
# an implementation of The AWK Programming Language, 1988.
# 
# Tested with Microsoft C vers 6 and 7, using the following make utils:
#   dmake-3.8, GNUish make, Kneller's Ndmake, and Microsoft's nmake

#========================== Configuration ===============================
# Compiler flags:
#  -AL large model
#  -Gs remove stack probes, -G2 requires 286-or-better
#  -O,-Ot optimize for time, -Os optimize for size
# MSDOS and _MSC_VER are predefined.
# MSDOS is not predifined in VC++ 1.5.

!ifdef EXE32
INCPATH=e:\msdev\include
LIBPATH=e:\msdev\lib
BINPATH=e:\msdev\bin
!ifndef debug
LIBS = libc.lib
OBJDIR=OBJN32
DEBUGOPT=
OPTFLAGS = -O1y -GBFs $(MATH)
CFLAGS = -nologo $(DEFS) $(OPTFLAGS)
LFLAGS= /NOLOGO /machine:I386 /MAP /DEBUG /DEBUGTYPE:BOTH \
 /subsystem:console /stack:256,64 /heap:1024,256
!else
OBJDIR=OBJD32
DEBUGOPT = -W3 -Zi -Od
CFLAGS = -nologo -GF  $(DEFS) $(DEBUGOPT)
LFLAGS= /NOLOGO /MACHINE:IX86 /MAP /subsystem:console
!endif
CC=$(BINPATH)\cl
LD=$(BINPATH)\link
O=.obj
YACC=kmyacc
!else
#!EXE32
INCPATH=E:\msvc\include
LIBPATH=E:\msvc\lib
BINPATH=E:\msvc\bin
#LIBS = llibc7
DEFS = -DMSDOS
STACK = 0x4800
!ifdef debug
CFLAGS = -nologo -AL -J -Zi -DDEBUG $(DEFS)
LFLAGS = /BATCH /CO /NOE /STACK:$(STACK) /ONERROR:NOEXE /MAP:FULL
OBJDIR = objd
!else ifdef fp87
CFLAGS = -nologo -AL -J -FPi87 -G3 $(DEFS)
LIBS = llibc7.lib
COPT = -Ox
OBJDIR = objn87
LFLAGS = /BATCH /STACK:$(STACK) /ONERROR:NOEXE /MAP:FULL
!else
CFLAGS = -nologo -AL -J -G3 $(DEFS)
LIBS = llibce.lib
COPT = -Ox
OBJDIR = objn
LFLAGS = /BATCH /NOE /STACK:$(STACK) /ONERROR:NOEXE /MAP:FULL
!endif
CC=$(BINPATH)\cl
LD=$(BINPATH)\link
YACC=kmyacc
!endif

#========================================================================


.c{$(OBJDIR)}.obj:
	@set include=$(INCPATH)
	$(CC) $(CFLAGS) $(COPT) -Fo$@ -c $<

OBJ1 = $(OBJDIR)\parse.obj $(OBJDIR)\array.obj $(OBJDIR)\bi_funct.obj \
       $(OBJDIR)\bi_vars.obj $(OBJDIR)\cast.obj 

OBJ2 = $(OBJDIR)\da.obj $(OBJDIR)\error.obj $(OBJDIR)\execute.obj \
       $(OBJDIR)\fcall.obj $(OBJDIR)\code.obj

OBJ3 = $(OBJDIR)\field.obj $(OBJDIR)\files.obj $(OBJDIR)\fin.obj \
       $(OBJDIR)\hash.obj $(OBJDIR)\jmp.obj $(OBJDIR)\init.obj

OBJ4 = $(OBJDIR)\kw.obj $(OBJDIR)\main.obj $(OBJDIR)\matherr.obj \
       $(OBJDIR)\memory.obj $(OBJDIR)\print.obj 

OBJ5 = $(OBJDIR)\re_cmpl.obj $(OBJDIR)\scan.obj $(OBJDIR)\scancode.obj \
       $(OBJDIR)\split.obj 

OBJ6 =  $(OBJDIR)\zmalloc.obj $(OBJDIR)\version.obj $(OBJDIR)\dosexec.obj

!ifdef EXE32
OBJ7 = $(OBJDIR)\xargs32.obj $(OBJDIR)\miscfunc.obj
!else
OBJ7 = $(OBJDIR)\xargs.obj $(OBJDIR)\miscfunc.obj
!endif

OBJ = $(OBJ1) $(OBJ2) $(OBJ3) $(OBJ4) $(OBJ5) $(OBJ6) $(OBJ7)

REXP_OBJ = $(OBJDIR)\rexp.obj $(OBJDIR)\rexp0.obj $(OBJDIR)\rexp1.obj \
           $(OBJDIR)\rexp2.obj $(OBJDIR)\rexp3.obj $(OBJDIR)\rexpdb.obj

RSP = $(OBJDIR)\mawk.lnk

all : $(OBJDIR) $(OBJDIR)\mawk.exe

$(OBJDIR)\mawk.exe : $(OBJ) $(REXP_OBJ)
	set LIB=$(LIBPATH)
        $(LD) @<<
!ifndef EXE32
$(LFLAGS)+
$(**: =+^
)
$@
$*.map
$(LIBS: =+^
)
$(DEF_FILE);
<<KEEP
!else
 $(LFLAGS) 
$(LIBS: =+^
)
/out:$(OBJDIR)\mawk32.exe
$(**: =^
)
<<KEEP
!endif

RFLAGS=-I. -Irexp -DMAWK

$(OBJDIR)\rexp.obj  :  rexp/rexp.c  rexp/rexp.h mbc.h
	@$(CC) $(CFLAGS) $(COPT) $(RFLAGS) -Fo$@ -c rexp/rexp.c

$(OBJDIR)\rexp0.obj  :  rexp/rexp0.c  rexp/rexp.h mbc.h
	@$(CC) $(CFLAGS) $(COPT) $(RFLAGS) -Fo$@ -c rexp/rexp0.c

$(OBJDIR)\rexp1.obj  :  rexp/rexp1.c  rexp/rexp.h mbc.h
	@$(CC) $(CFLAGS) $(COPT) $(RFLAGS) -Fo$@ -c rexp/rexp1.c

$(OBJDIR)\rexp2.obj  :  rexp/rexp2.c  rexp/rexp.h mbc.h
	@$(CC) $(CFLAGS) $(COPT) $(RFLAGS) -Fo$@ -c rexp/rexp2.c

$(OBJDIR)\rexp3.obj  :  rexp/rexp3.c  rexp/rexp.h mbc.h
	@$(CC) $(CFLAGS) $(COPT) $(RFLAGS) -Fo$@ -c rexp/rexp3.c

$(OBJDIR)\rexpdb.obj  :  rexp/rexpdb.c  rexp/rexp.h mbc.h
	@$(CC) $(CFLAGS) $(COPT) $(RFLAGS) -Fo$@ -c rexp/rexpdb.c

config.h : msdos/msc.h
	copy msdos\msc.h  config.h
	copy msdos\mawk.def .

dosexec.c : msdos/dosexec.c
	copy msdos\dosexec.c dosexec.c

test : mawk.exe  # test that we have a sane mawk
	@echo you may have to run the test manually
	cd test && mawktest.bat

fpe_test :  mawk.exe # test FPEs are handled OK
	@echo testing floating point exception handling
	@echo you may have to run the test manually
	cd test && fpe_test.bat

###################################################
# parse.c is provided 
# so you don't need to make it.
#
# But if you do:  here's how:
# To make it with bison under msdos
#YACC=bison -y
parse.c : parse.y 
	$(YACC) -d parse.y
	mv ytab.h parse.h
	mv ytab.c parse.c
########################################

#scancode.c :  makescan.c  scan.h
#	$(CC) -o makescan.exe  makescan.c
#	makescan.exe > scancode.c
#	del makescan.exe

clean :
	@del $(OBJDIR)\*.obj

distclean :
	@del $(OBJDIR)\*.obj
	@del config.h dosexec.c
	@del $(OBJDIR)\mawk.exe

COMMONH = config.h mawk.h nstd.h sizes.h types.h
MEMH = memory.h zmalloc.h

#  dependencies of .objs on .h
$(OBJDIR)\array.obj : $(COMMONH) $(MEMH) field.h bi_vars.h symtype.h
$(OBJDIR)\bi_funct.obj : $(COMMONH) $(MEMH) field.h bi_vars.h init.h regexp.h symtype.h repl.h bi_funct.h files.h fin.h mbc.h
$(OBJDIR)\bi_vars.obj : $(COMMONH) $(MEMH) field.h bi_vars.h init.h symtype.h
$(OBJDIR)\code.obj : $(COMMONH) $(MEMH) field.h code.h init.h symtype.h jmp.h
$(OBJDIR)\da.obj : $(COMMONH) $(MEMH) field.h code.h symtype.h repl.h bi_funct.h
$(OBJDIR)\error.obj : $(COMMONH) bi_vars.h parse.h vargs.h symtype.h scan.h
$(OBJDIR)\execute.obj : $(COMMONH) $(MEMH) field.h bi_vars.h code.h regexp.h symtype.h repl.h bi_funct.h fin.h
$(OBJDIR)\fcall.obj : $(COMMONH) $(MEMH) code.h symtype.h
$(OBJDIR)\field.obj : $(COMMONH) $(MEMH) field.h bi_vars.h init.h parse.h regexp.h symtype.h repl.h scan.h
$(OBJDIR)\files.obj : $(COMMONH) $(MEMH) files.h fin.h
$(OBJDIR)\fin.obj : $(COMMONH) $(MEMH) field.h bi_vars.h parse.h symtype.h scan.h fin.h
$(OBJDIR)\hash.obj : $(COMMONH) $(MEMH) symtype.h
$(OBJDIR)\init.obj : $(COMMONH) $(MEMH) field.h bi_vars.h code.h init.h symtype.h xargs.h
$(OBJDIR)\jmp.obj : $(COMMONH) $(MEMH) code.h init.h symtype.h jmp.h
$(OBJDIR)\kw.obj : $(COMMONH) init.h parse.h symtype.h
$(OBJDIR)\main.obj : $(COMMONH) $(MEMH) field.h bi_vars.h code.h init.h symtype.h files.h fin.h
$(OBJDIR)\makescan.obj : parse.h symtype.h scan.h
$(OBJDIR)\matherr.obj : $(COMMONH)
$(OBJDIR)\memory.obj : $(COMMONH) $(MEMH)
$(OBJDIR)\parse.obj : $(COMMONH) $(MEMH) field.h bi_vars.h code.h symtype.h bi_funct.h files.h jmp.h
$(OBJDIR)\print.obj : $(COMMONH) $(MEMH) field.h bi_vars.h parse.h symtype.h scan.h bi_funct.h files.h
$(OBJDIR)\re_cmpl.obj : $(COMMONH) $(MEMH) parse.h regexp.h symtype.h repl.h scan.h
$(OBJDIR)\scan.obj : $(COMMONH) $(MEMH) field.h code.h init.h parse.h symtype.h repl.h scan.h files.h fin.h
$(OBJDIR)\split.obj : $(COMMONH) $(MEMH) field.h bi_vars.h parse.h regexp.h symtype.h scan.h bi_funct.h mbc.h
$(OBJDIR)\version.obj : $(COMMONH) patchlev.h
$(OBJDIR)\zmalloc.obj : $(COMMONH) zmalloc.h

$(OBJDIR)\miscfunc.obj : miscfunc.c

$(OBJDIR)\xargs.obj : xargs.c xargs.h
	$(CC) $(CFLAGS)  $(COPT) -DLARGE_DATA -Fo$@ -c xargs.c

$(OBJDIR)\cast.obj : $(COMMONH) $(MEMH) field.h parse.h symtype.h repl.h scan.h
	$(CC) $(CFLAGS)  -O -Fo$@ -c cast.c

$(OBJDIR)\init.obj : $(COMMONH) $(MEMH) field.h bi_vars.h code.h init.h symtype.h xargs.h
	$(CC) $(CFLAGS) -O -Fo$@ -c init.c

##########
$(OBJDIR):
	@if not exist $(OBJDIR)\NUL md $(OBJDIR)
