/* Id: xargs.c 2.3 1993/09/04 17:09:31 serow Exp serow 
 *
 * wild card expand routine for BORLAND C, MS C, LSI C..
 * This code is in the public domain.
 *
 * $Log: xargs32.c $
 * Revision 1.2  1997/07/18  17:13:11  koichik
 * 32bitΉ̔ɑΏ
 *
 * Revision 1.1  1997/04/27  15:01:25  koichik
 * Initial revision
 *
 * Revision 2.3  1993/09/04  17:09:31  serow
 * _heapmin (MS-C)
 *
 * Revision 2.2  1993/05/22  04:33:02  serow
 * Oargv[0]
 *
 * Revision 2.1  1993/01/20  10:43:04  serow
 * v2
 *
 */

#include <assert.h>
#include <stdio.h>
#include <dos.h>
#include <string.h>
#include <io.h>
#include <stdlib.h>
#include <windows.h>
#include <winbase.h>
#include "xargs32.h"

#ifdef TEST
#define static
#endif

static unsigned long xargs_rule = XARGS_CSH_LIKE_RULE;

#define OnlySubDir  (1<<31)
#define AllFindAttr (0xB7)
#define INVERTCHAR  '^'
#define LBRACE      '['
#define RBRACE      ']'
#define LCURLY      '{'
#define RCURLY      '}'
#define Isend(c)    ((c) == '\0' || ((xargs_rule & XARGS_DOTBREAK) && (c) == '.'))
#define Iswild(c)   ((c) == '*' || (c) == '?' || (c) == LBRACE || (c) == LCURLY)
#define Isupper(c)  ('A' <= (c) && (c) <= 'Z')
#define Isspace(c)  ((c) == ' ' || (c) == '\t' || (c) == '\n' || (c) == '\r')
/* ascii only */
#define Tolower(c)  ((c) | 0x20)
/* Shift-JIS only */
#define isk1(c) ((0x81 <= (unsigned char)(c) && (unsigned char)(c) <= 0x9F)\
	      || (0xE0 <= (unsigned char)(c) && (unsigned char)(c) <= 0xFC))

#define REALLOC_UNIT    512

static char ERNOMEM[] = "not enought memory.\r\n";

/***************************************************************/

static void
error(char * s)
{
	write(2, "xargs: ", 7);
	write(2, s, strlen(s));
	exit(1);
}

static char *
find_file(char *dir, HANDLE *hFind, DWORD Attr, WIN32_FIND_DATA *dta)
{
    char *ret = NULL;

    assert(hFind != NULL);

    if (*hFind == INVALID_HANDLE_VALUE) {      /* get first entry */
	dta->dwFileAttributes = (DWORD)(Attr & 0xffff);
	if ((*hFind = FindFirstFile(dir, dta)) != INVALID_HANDLE_VALUE)
	    ret = dta->cFileName;
    } else {                    /* get next entry */
retry:
	ret = NULL;
	if (FindNextFile(*hFind, dta))
	    ret = dta->cFileName;
    }

    if (ret == NULL)
	return NULL;

    if (!strcmp(ret, "..") || !strcmp(ret, "."))
	goto retry;

    if ((Attr & OnlySubDir) && (dta->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
	goto retry;

    if (dta->cFileName[0] == '.' && (xargs_rule & XARGS_HIDDENDOT) 
	&& (dta->dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) == 0)
	goto retry;

    return ret;
}

static char *
jstrtolwr(char *p)
{
    char *op = p;

    while (*p) {
	if (isk1(*p) && p[1])
	    p++;
	else if (Isupper(*p))
	    *p = (char) Tolower(*p);
	p++;
    }
    return op;
}

static char *
jrindex(char *p, int c)
{
    char *oldp;

    c &= 0xFF;
    for (oldp = 0; *p; p++) {
	if (isk1(*p) && p[1]) {
	    p++;
	} else if (((*p) & 0xFF) == c) {
	    oldp = p;
	}
    }
    return oldp;
}

static int
ustrcmp(void const *s1, void const *s2)
{
    unsigned char *p1 = *(unsigned char **)s1;
    unsigned char *p2 = *(unsigned char **)s2;

    while ((*p1 != '\0') && (*p1 == *p2)) {
	p1++;
	p2++;
    }
    return *p1 - *p2;
}

static char *
xmalloc(size_t n)
{
    char *bp;

    if ((bp = calloc(1, n)) == (char *) 0)
	error(ERNOMEM);
    return bp;
}

static char *
xrealloc(void *blk, size_t n)
{
    char *bp;

    if ((bp = realloc(blk, n)) == (char *) 0)
	error(ERNOMEM);

    return bp;
}
static void
xfree(void *p)
{
    if (p)
	free(p);

}

static int
has_glob_char(char *pat)
{
    int c;

    while ((c = *pat++) != '\0') {
	if (Iswild(c))
	    return 1;
	if (isk1(c) && *pat)
	    pat++;
    }
    return 0;
}

#define Scan(c, p) {\
	c = (unsigned char)(*p++);\
	if (isk1(c) && *p) \
	    c = (c << 8) + (unsigned char)(*p++);\
	else if (Isupper(c))\
	    c = Tolower(c);\
	}
#define ScanRaw(c, p) {\
	c = (unsigned char)(*p++);\
	if (isk1(c) && *p)\
	    c = (c << 8) + (unsigned char)(*p++);\
	}

static int
glob_match(char *text, char *pattern)
{
    int include, exclude;
    unsigned int ks, ke, kp, kt;
    char *start, *stop;

    while (*pattern) {
	switch (*pattern) {
	case LCURLY:
	    if (Isend(*text))
		return 0;
	    stop = start = text;
	    pattern++;
	    do {
		text = start;
		include = 1;
		do {
		    if (Isend(*pattern))
			return 0; /*ERMISSCURLY*/
		    Scan(kt, text);
		    Scan(kp, pattern);
		    if (kt != kp)
			include = 0;
		} while (*pattern != ',' && *pattern != RCURLY);

		if (include && text - start > stop - start)
		    stop = text;
		if (*pattern == ',')
		    pattern++;
	    } while (*pattern != RCURLY);
	    if (stop == start)
		return 0;
	    pattern++;
	    text = stop;
	    break;

	case '?':
	    if (Isend(*text))
		return 0;
	    ScanRaw(kt, text);
	    pattern++;
	    break;

	case '*':
	    while ((include = glob_match(text, pattern + 1)) == 0 && !Isend(*text)) {
		ScanRaw(kt, text);
	    }
	    return include;

	case LBRACE:
	    if (Isend(*text))
		return 0;
	    include = exclude = 0;
	    pattern++;
	    if (*pattern == INVERTCHAR) {
		exclude = 1;
		pattern++;
	    }
	    Scan(kt, text);
	    do {
		if (Isend(*pattern))
		    return 0; /*ERMISSBRACKET*/
		Scan(ks, pattern);
		ke = ks;
		if (*pattern == '-' && *(pattern + 1) != RBRACE) {
		    pattern++;
		    if (Isend(*pattern))
			return 0; /*ERMISSBRACKET*/
		    Scan(ke, pattern);
		}
		if (ks <= kt && kt <= ke) {
		    include = 1;
		    /* no break to checking missing brace */
		    /*break;*/
		}
	    } while (*pattern != RBRACE);

	    if (include == exclude)
		return 0;
	    pattern++;
	    break;

	case '.':
	    if (xargs_rule & XARGS_DOTBREAK) {
		if (!Isend(*text))
		    return 0;
		if (*text == '.')
		    text++;
		pattern++;
		break;
	    }
	default:
	    Scan(kt, text);
	    Scan(kp, pattern);
	    if (kt != kp)
		return 0;
	}
    }
    return *text == '\0' ? 1 : 0;
}

static char *
jrpathdlm(char *s)
{
    char *ap, *ap1, *ap2;

    ap1 = jrindex(s, '\\');
    ap2 = jrindex(s, '/');
    if ((ap = (ap1 < ap2) ? ap2 : ap1) == (char *) 0)
	ap = jrindex(s, ':');
    return ap;
}

char **
glob_filename(char *pattern)
{
    char *ap, *ap1, *ap2;
    char **res;
    char *pat;
    char **paths;
    char *dirp;
    char *wld;
    char *pp;
    int ra, ri, i, plen;
    int np;
    DWORD Attr = ((DWORD) (xargs_rule & 0xFFFF));

    HANDLE hFind = INVALID_HANDLE_VALUE;
    WIN32_FIND_DATA dta;

    ap = jrpathdlm(pattern);
    if (ap && ap[0] != ':' && ap[1] == '\0') {
		Attr = AllFindAttr | OnlySubDir;
	pp = (char *) xmalloc(strlen(pattern) + 1);
	strcpy(pp, pattern);
	pp[(int) (ap - pattern)] = '\0';
	ap = jrpathdlm(pp);
	if (ap)
	    ap = pattern + (int) (ap - pp);
	xfree(pp);
    }
    if (ap) {
	pat = xmalloc(strlen(ap + 1) + 1);
	strcpy(pat, ap + 1);
	ap2 = dirp = (char *) xmalloc((int) (ap - pattern) + 2);
	for (ap1 = pattern; ap1 <= ap; ap1++)
	    *ap2++ = *ap1;
	*ap2 = '\0';
	if (ap[0] != ':' && has_glob_char(dirp)) {
	    paths = glob_filename(dirp);
	    xfree(dirp);
	} else {
	    paths = (char **) xmalloc(sizeof(paths[0]) * 2);
	    paths[0] = dirp;
	    paths[1] = (char *) 0;
	}
    } else {
	pat = xmalloc(strlen(pattern) + 1);
	strcpy(pat, pattern);
	paths = (char **) xmalloc(sizeof(paths[0]) * 2);
	paths[0] = (char *) xmalloc(2);
	paths[1] = (char *) 0;
	strcpy(paths[0], "");
    }
    ap = jrpathdlm(pat);
    if (ap && ap[0] != ':' && ap[1] == '\0') {
	ap[0] = '\0';
    }
    res = (char **) xmalloc(sizeof(char *) * (ra = 260));
    ri = 0;
    res[ri] = 0;
    for (i = 0; paths[i]; i++) {
	plen = strlen(paths[i]);
	wld = xmalloc(plen + 5);
	strcpy(wld, paths[i]);
	ap = jrpathdlm(wld);

	if (plen == 0 || (ap && ap[1] == '\0')) {
	    np = 0;
	} else {
	    strcat(wld, "/");
	    np = 1;
	}

	strcat(wld, "*.*");

	if ((ap = find_file(wld, &hFind, Attr, &dta)) != NULL) {
	    do {
		if (glob_match(ap, pat)) {
		    if (xargs_rule & XARGS_CVLOWER)
			jstrtolwr(ap);
		    res[ri] = xmalloc(plen + strlen(ap) + 2);
		    strcat(strcat(strcpy(res[ri], paths[i]), np ? "/" : ""), ap);
		    res[++ri] = (char *) 0;
		    if (ri >= ra - 2) {
			    res = (char **) xrealloc(res, sizeof(char *) * (ra += 260));
		    }
		}
	    } while (find_file(wld, &hFind, Attr, &dta));

	    FindClose(hFind);
	    hFind=INVALID_HANDLE_VALUE;
	}

	xfree(wld);
    }
    if (ri) {
	if (xargs_rule & XARGS_DOSORT)
	    qsort(&res[0], ri, sizeof(res[0]), ustrcmp);
    } else {
	strcpy(res[ri] = xmalloc(strlen(pattern) + 1), pattern);
	res[++ri] = (char *) 0;
    }
    xfree(pat);
    for (i = 0; paths[i]; i++)
	xfree(paths[i]);
    xfree(paths);
    return res;
}

/******/
struct inputredir {
    int type;
#define IP_FILE 1
#define IP_MEM  2
    struct inputredir *next;
    union {
	FILE *fp;
	char *cp;
    } i;
};
static struct inputredir *iroot = 0;

static void
add_input(void *p, int type)
{
    struct inputredir *ip;

    ip = (struct inputredir *) xmalloc(sizeof(struct inputredir));
    ip->type = type;
    ip->i.cp = p;
    if (iroot) {
	ip->next = iroot;
	iroot = ip;
    } else {
	iroot = ip;
	iroot->next = 0;
    }
}

static int pushbackc = -1;

static int
nextc(void)
{
    int c;
    struct inputredir *lip;

    if (pushbackc != -1) {
	c = pushbackc;
	pushbackc = -1;
	return c;
    }
    if (iroot == 0)
	return EOF;
    switch (iroot->type) {
    case IP_FILE:
	if ((c = fgetc(iroot->i.fp)) == EOF) {
	    fclose(iroot->i.fp);
	    lip = iroot;
	    iroot = iroot->next;
	    xfree(lip);
	    c = nextc();
	}
	break;
    case IP_MEM:
	if ((c = *(iroot->i.cp++)) == '\0') {
	    lip = iroot;
	    iroot = iroot->next;
	    xfree(lip);
	    c = nextc();
	}
	break;
    default:
	c = EOF;
	break;
    }
    return c;
}
static void
pushback(int c)
{
    pushbackc = c;
}

/******/

static void
setav(char *NCptr, int *_ARGC, char ***_ARGV)
{
    int   c, i, jf, at;
    int   quote, withwild;
    char *ap;
    char *buf;
    size_t absiz;
    size_t bfsiz;
#define argbuf (*_ARGV)
#define argcnt (*_ARGC)
    char **res;
    FILE  *fp;

    buf = (char *) xmalloc((bfsiz = REALLOC_UNIT * 2) * sizeof(char));
    argbuf = (char **) xmalloc((absiz = REALLOC_UNIT) * sizeof(char *));
    argcnt = 0;

    add_input(NCptr, IP_MEM);

    while ((c = nextc()) != EOF) {
	withwild = quote = 0;
	while (Isspace(c))
	    c = nextc();
	if (c == EOF)
	    break;
	at = (c == '@') ? 1 : 0;
	ap = buf;
	jf = 0;
	do {
	    if (ap == &buf[bfsiz - 1]) {
		buf = (char *) xrealloc(buf, (bfsiz + REALLOC_UNIT) * sizeof(buf[0]));
		ap = &buf[bfsiz - 1];
		bfsiz += REALLOC_UNIT;
	    }
	    if (jf || isk1(c)) {
		*ap++ = (char) c;
		jf = !jf;
		continue;
	    }
	    if (quote !='\'' && Iswild(c))
		withwild = 1;
	    if (c == quote) {   /* end of quoted string */
		quote = 0;
	    } else if (!quote && (c == '\'' || c == '"')) {
		quote = c;
	    } else {
		if (c == '\\') {
		    c = nextc();
		    if (c != '"' && c != '\'') {
			pushback(c);
			c = '\\';
		    }
		}
		*ap++ = (char) c;
	    }
	} while ((c = nextc()) != EOF && (quote || !Isspace(c)));
	*ap = 0;
	if ((xargs_rule & XARGS_ATRESP) && at) {
	    if ((fp = fopen(buf + 1, "r")) != 0) {
		add_input(fp, IP_FILE);
		continue;
	    } else {
		/* ignore */
	    }
	}
	if (withwild) {
	    res = glob_filename(buf);
	} else {
	    res = (char **) xmalloc(sizeof(char *) * 2);
	    strcpy(res[0] = (char *) xmalloc(strlen(buf) + 1), buf);
	    res[1] = (char *) 0;
	}
	for (i = 0; res[i]; i++) {
	    argbuf[argcnt++] = res[i];
	    if ((size_t) argcnt >= absiz) {
		argbuf = (char **) xrealloc(argbuf, (absiz += REALLOC_UNIT) * sizeof(argbuf[0]));
	    }
	}
	xfree(res);
    }
    argbuf[argcnt] = (char *) 0;
    argbuf = (char **) xrealloc(argbuf, (argcnt + 1) * sizeof(argbuf[0]));
    xfree(buf);
}

unsigned long
xargs_set_rule(unsigned long rule)
{
    unsigned long lastrule;

    lastrule = xargs_rule; 
    xargs_rule = rule; 
    return lastrule;
}

unsigned long
xargs_modify_rule(unsigned long mask, unsigned long rule)
{
    unsigned long lastrule;

    lastrule = xargs_rule; 
    xargs_rule = (xargs_rule & ~mask) | rule;
    return lastrule;
}

unsigned char cmdlinebuf[1024];

void
xargs(int *argc, char ***argv)
{
    char far *pp;
    char *s;
    int i, j;
    int jf = 0;
    int li = 0;
    int sa = 0;

    if (!(xargs_rule & XARGS_USECMDL)
      || (s = getenv("CMDLINE")) == 0
      || getenv("NOCMDLINE") != 0) {
	sa = 1;
	pp = "xargs";
	s = xmalloc(1024);
	
#if 0        
	{
	    cmdlinebuf[0] = '\0';
	    //wcstombs(cmdlinebuf, GetCommandLine(), (sizeof (cmdlinebuf) - 1));
	    //pp = cmdlinebuf;
	    strcpy(cmdlinebuf, GetCommandLine());
	    pp = cmdlinebuf;
	}

	for (i = 0; (s[i] = *pp++) != '\0'; i++) {
	    //if (isspace(s[i])) {
	    //    s[i] = '\0';
	    //    break;
	    //}
	    
	    if (jf) {
		jf = 0;
		continue;
	    }
	    if (isk1(s[i]) && !jf) {
		jf = 1;
		continue;
	    }
	    if ((xargs_rule & XARGS_0CVLOWER) && Isupper(s[i]))
		s[i] = (char) Tolower(s[i]);
	    if (s[i] == '.')
		li = i;
	    if (s[i] == '/' || s[i] == '\\')
		li = 0;
	    if ((xargs_rule & XARGS_0CVSLASH) && s[i] == '\\')
		s[i] = '/';
	}
	if ((xargs_rule & XARGS_0REMOVEXT) && li && li != i - 1) {
	    i = li;
	    s[i] = 0;
	}
	{
	    //cmdlinebuf[0] = '\0';
	    //wcstombs(cmdlinebuf, GetCommandLine(), (sizeof (cmdlinebuf) - 1));
	    //pp = cmdlinebuf;
	    //strcpy(cmdlinebuf, GetCommandLine());
	    // pp = cmdlinebuf;
	    //while (*pp && !isspace(*pp))
	    //    *pp++;
	    //j = *pp;
	    pp = cmdlinebuf+i;
	    j = strlen(cmdlinebuf+i);

	}
	while ((*pp == ' ' || *pp == '\t') && j) {
	    pp++;
	    j--;
	}
	if (j)
	    s[i++] = ' ';
	while (j--) {
	    s[i++] = *pp++;
	}
	s[i] = 0;
#else    
	s[0] = '\0';
	strcpy(s, GetCommandLine());
#endif
    }
    setav(s, argc, argv);
    if (sa)
	xfree(s);
}


#ifdef TEST
#include <stdlib.h>

static char QUOTE[]= "\033[33m\"\033[m";

int
main(int argc, char **argv)
{
    int a = 0;

    /*xargs_modify_rule(XARGS_DOTBREAK, XARGS_DOTBREAK);*/
    xargs(&argc, &argv);

    printf("[argv]-------------------------- %d argument(s)\n", argc);
    while (argc--) {
	printf("[%4d]%s%s%s\n", a++, QUOTE, *argv++, QUOTE);
    }
    printf("-------------------------------- Thats all\n");

	return 0;
}

#endif
